#include <stdafx.h>
#include "complex.h"
#include "problem.h"
#include "fullmatrix.h"

/////////////////////////////////////////////////////////////////////////////
// CNode construction

CNode::CNode()
{
	x=0.;
	y=0.;
	IsSelected=FALSE;
	BoundaryMarker=-1;
}

CComplex CNode::CC()
{
	return (x+I*y);
}

double CNode::GetDistance(double xo, double yo)
{
	return sqrt((x-xo)*(x-xo) + (y-yo)*(y-yo));
}

void CNode::ToggleSelect()
{
	if (IsSelected==TRUE) IsSelected=FALSE;
	else IsSelected=TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CMeshNode construction

CMeshNode::CMeshNode()
{
	x=0.; y=0.;
	A.re=0; A.im=0;
}

CComplex CMeshNode::CC()
{
	return (x+I*y);
}

double CMeshNode::GetDistance(double xo, double yo)
{
	return sqrt((x-xo)*(x-xo) + (y-yo)*(y-yo));
}

/////////////////////////////////////////////////////////////////////////////
// CSegment construction

CSegment::CSegment()
{
	n0=0;
	n1=0;
	IsSelected=FALSE;
	BoundaryMarker=-1;
}

void CSegment::ToggleSelect()
{
	if (IsSelected==TRUE) IsSelected=FALSE;
	else IsSelected=TRUE;
}
/////////////////////////////////////////////////////////////////////////////
// CArcSegment construction

CArcSegment::CArcSegment()
{
	n0=0;
	n1=0;
	IsSelected=FALSE;
	MaxSideLength=-1;
	ArcLength=90.;
	BoundaryMarker=-1;
}

void CArcSegment::ToggleSelect()
{
	if (IsSelected==TRUE) IsSelected=FALSE;
	else IsSelected=TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CNode construction

CBlockLabel::CBlockLabel()
{
	x=0.;
	y=0.;
	MaxArea=0.;
	IsSelected=FALSE;
	BlockType=-1;
}

void CBlockLabel::ToggleSelect()
{
	if (IsSelected==TRUE) IsSelected=FALSE;
	else IsSelected=TRUE;
}

double CBlockLabel::GetDistance(double xo, double yo)
{
	return sqrt((x-xo)*(x-xo) + (y-yo)*(y-yo));
}

CMaterialProp::CMaterialProp()
{		
		BlockName="New Material";
		mu_x=1.;
		mu_y=1.;			// permeabilities, relative
		BHpoints=0;		
		BHdata=NULL;
		slope=NULL;
		H_c=0.;				// magnetization, A/m
		Jr=0.;
		Ji=0.;				// applied current density, MA/m^2
		Cduct=0.;		    // conductivity of the material, MS/m
		Lam_d=0.;			// lamination thickness, mm
		Theta_h=0.;			// hysteresis angle, degrees
		LamFill=1.;			// lamination fill factor;
		LamType=0;			// type of lamination;
}

CMaterialProp::~CMaterialProp()
{
	if (BHdata!=NULL)	free(BHdata);
	if (slope!=NULL)	free(slope);
}

void CMaterialProp::GetSlopes()
{
	if (BHpoints==0) return; // catch trivial case;

	CFullMatrix Z;
	int i;
	double l1,l2;
	double *B,*H;
	double uo=4.e-07*PI;

	// take care of LamType=0 situation by changing the
	// apparent B-H curve;
	if(LamType==0)
	{
		for(i=0;i<BHpoints;i++){
			BHdata[i].re = LamFill*BHdata[i].re + 
				           (1.-LamFill)*uo*BHdata[i].im;
		}
	}

	Z.Create(BHpoints);
	Z.Wipe();
	slope=(double *)calloc(BHpoints,sizeof(double));
	B=(double *)calloc(BHpoints,sizeof(double));
	H=(double *)calloc(BHpoints,sizeof(double));
	for(i=0;i<BHpoints;i++){
		B[i]=BHdata[i].re;
		H[i]=BHdata[i].im;
	}
	
	// smooth out input points to get rid of noise in entered points.
	for(i=1;i<BHpoints-1;i++){
        l1=B[i]-B[i-1];
        l2=B[i+1]-B[i];
        slope[i]=(l2*l2*(H[i-1] + H[i]) +
            l1*l2*(H[i-1]+H[i+1]) +
            l1*l1*(H[i]+H[i+1]))/
            (2*(l1*l1 + l1*l2 + l2*l2));
	}
	for(i=1;i<BHpoints-1;i++) H[i]=slope[i];

	// impose natural BC on the `left'
	l1=B[1]-B[0];
	Z.M[0][0]=4./l1;
	Z.M[0][1]=2./l1;
	Z.b[0]=6.*(H[1]-H[0])/(l1*l1);

	// impost natural BC on the `right'
	int n=BHpoints;
	l1=B[n-1]-B[n-2];
	Z.M[n-1][n-1]=4./l1;
	Z.M[n-1][n-2]=2./l1;
	Z.b[n-1]=6.*(H[n-1]-H[n-2])/(l1*l1);

	for(i=1;i<BHpoints-1;i++)
	{
		l1=B[i]-B[i-1];
        l2=B[i+1]-B[i];

		Z.M[i][i-1]=2./l1;
		Z.M[i][i]=4.*(l1+l2)/(l1*l2);
		Z.M[i][i+1]=2./l2;

		Z.b[i]=6.*(H[i]-H[i-1])/(l1*l1) +
				 6.*(H[i+1]-H[i])/(l2*l2);
	}

	Z.GaussSolve();

	for(i=0;i<BHpoints;i++) slope[i]=Z.b[i];
    free(B);
	free(H);

	return;
}


double CMaterialProp::GetH(double x)
{
	double b,h,z,z2,l;
	int i;

	b=fabs(x);
	
	if(BHpoints==0)	return 0;

	if(b>BHdata[BHpoints-1].re)
		return (BHdata[BHpoints-1].im + 
		        slope[BHpoints-1]*(b-BHdata[BHpoints-1].re));

	for(i=0;i<BHpoints-1;i++)
		if((b>=BHdata[i].re) && (b<=BHdata[i+1].re)){
			l=cRe(BHdata[i+1]-BHdata[i]);
			z=(b-BHdata[i].re)/l;
			z2=z*z;
			h=(1.-3.*z2+2.*z2*z)*BHdata[i].im +
			  z*(1.-2.*z+z2)*l*slope[i] +
			  z2*(3.-2.*z)*BHdata[i+1].im +
			  z2*(z-1.)*l*slope[i+1];
			return h;
		}

	return 0;
}

double CMaterialProp::GetEnergy(double x)
{
	double b,z,z2,l,nrg;
	double b0,b1,h0,h1,dh0,dh1;
	int i;

	b=fabs(x);
	nrg=0.;
	
	if(BHpoints==0)	return 0;

	for(i=0;i<BHpoints-1;i++){

		b0=BHdata[i].re;	h0=BHdata[i].im;
		b1=BHdata[i+1].re;	h1=BHdata[i+1].im;
		dh0=slope[i];
		dh1=slope[i+1];

		if((b>=b0) && (b<=b1)){
			l=b1-b0;
			z=(b-b0)/l;
			z2=z*z;

			nrg += (dh0*l*l*(6. + z*(-8. + 3.*z))*z2)/12. + 
				(h0*l*z*(2. + (-2. + z)*z2))/2. - 
				(h1*l*(-2. + z)*z2*z)/2. +			
				(dh1*l*l*(-4. + 3.*z)*z2*z)/12;

			return nrg;
		}
		else{
			// point isn't in this section, but add in the
			// energy required to pass through it.
			b0=BHdata[i].re;	h0=BHdata[i].im;
			b1=BHdata[i+1].re;	h1=BHdata[i+1].im;
			dh0=slope[i];
			dh1=slope[i+1];
			nrg += ((b0 - b1)*((b0 - b1)*(dh0 - dh1) -
				6.*(h0 + h1)))/12.;
		}
	}		

	// if we've gotten to this point, the point is off the scale,
	// so we have to extrapolate the rest of the way...

	h0=BHdata[BHpoints-1].im;
	dh0=slope[BHpoints-1];
	b0=BHdata[BHpoints-1].re;

	nrg += ((b - b0)*(b*dh0 - b0*dh0 + 2*h0))/2.;
	
	return nrg;
}

double CMaterialProp::GetCoEnergy(double b)
{
	return (fabs(b)*GetH(b) - GetEnergy(b));
}

double CMaterialProp::DoEnergy(double b1, double b2)
{
	// calls the raw routine to get point energy, 
	// but deals with the load of special cases that
	// arise because I insisted on trying to deal with
	// laminations on a continuum basis.

	double uo=4.e-07*PI;
	double nrg,biron,bair;

	// easiest case: the material is linear!
	if (BHpoints==0)
	{
		double h1,h2;
		
		if(LamType==0){		// laminated in-plane
			h1=b1/((1.+LamFill*(mu_x-1.))*uo);
			h2=b2/((1.+LamFill*(mu_y-1.))*uo);
		}

		if(LamType==1){		// laminated parallel to x;
			h1=b1/((1.+LamFill*(mu_x-1.))*uo);
			h2=b1*(LamFill/(mu_y*uo) + (1. - LamFill)/uo);
		}

		if(LamType==2){		// laminated parallel to x;
			h2=b1/((1.+LamFill*(mu_y-1.))*uo);
			h1=b1*(LamFill/(mu_x*uo) + (1. - LamFill)/uo);
		}

		return ((h1*b1+h2*b2)/2.);
	}

	// Rats! The material is nonlinear.  Now, we have to do
	// a bit of work to get the energy.
	if(LamType==0) nrg = GetEnergy(sqrt(b1*b1+b2*b2));

	if(LamType==1){
		biron=sqrt((b1/LamFill)*(b1/LamFill) + b2*b2);
		bair=b2;
		nrg = LamFill*GetEnergy(biron)+(1-LamFill)*bair*bair/(2.*uo);
	}

	if(LamType==2){
		biron=sqrt((b2/LamFill)*(b2/LamFill) + b1*b1);
		bair=b1;
		nrg = LamFill*GetEnergy(biron)+(1-LamFill)*bair*bair/(2.*uo);
	}

	return nrg;
}

double CMaterialProp::DoCoEnergy(double b1, double b2)
{
	double uo=4.e-07*PI;
	double nrg,biron,bair;

	// easiest case: the material is linear!
	// in this case, energy and coenergy are the same!
	if (BHpoints==0) return DoEnergy(b1,b2);

	if(LamType==0) nrg = GetCoEnergy(sqrt(b1*b1+b2*b2));

	if(LamType==1){
		biron=sqrt((b1/LamFill)*(b1/LamFill) + b2*b2);
		bair=b2;
		nrg = LamFill*GetCoEnergy(biron)+(1-LamFill)*bair*bair/(2.*uo);
	}

	if(LamType==2){
		biron=sqrt((b2/LamFill)*(b2/LamFill) + b1*b1);
		bair=b1;
		nrg = LamFill*GetCoEnergy(biron)+(1-LamFill)*bair*bair/(2.*uo);
	}

	return nrg;
}

void CMaterialProp::GetMu(double b1, double b2,
							double &mu1, double &mu2)
{
	// gets the permeability, given a flux density
	//

	double uo=4.e-07*PI;
	double biron;

	// easiest case: the material is linear!
	if (BHpoints==0)
	{		
		if(LamType==0){		// laminated in-plane
			mu1=((1.+LamFill*(mu_x-1.))*uo);
			mu2=((1.+LamFill*(mu_y-1.))*uo);
		}

		if(LamType==1){		// laminated parallel to x;
			mu1=((1.+LamFill*(mu_x-1.))*uo);
			mu2=1./(LamFill/(mu_y*uo) + (1. - LamFill)/uo);
		}

		if(LamType==2){		// laminated parallel to x;
			mu2=((1.+LamFill*(mu_y-1.))*uo);
			mu1=1./(LamFill/(mu_x*uo) + (1. - LamFill)/uo);
		}
	}

	// Rats! The material is nonlinear.  Now, we have to do
	// a bit of work to get the energy.
else{
	double muiron;

	if(LamType==0){
		biron=sqrt(b1*b1+b2*b2);
		if(biron<1.e-08) muiron=slope[0]; //catch degenerate case
		else mu1=biron/GetH(biron);
		mu2=mu1;
	}

	if(LamType==1){
		biron=sqrt((b1/LamFill)*(b1/LamFill) + b2*b2);
		if(biron<1.e-08) muiron=slope[0];
		else muiron=biron/GetH(biron);
		mu1=muiron*LamFill;
		mu2=1./(LamFill/muiron + (1. - LamFill)/uo);
		
	}

	if(LamType==2){
		biron=sqrt((b2/LamFill)*(b2/LamFill) + b1*b1);
		if(biron<1.e-08) muiron=slope[0];
		else muiron=biron/GetH(biron);
		mu2=muiron*LamFill;
		mu1=1./(LamFill/muiron + (1. - LamFill)/uo);
	}
}
	// convert to relative permeability
	mu1/=uo;
	mu2/=uo;

	return;
}









CBoundaryProp::CBoundaryProp()
{
		BdryName="New Boundary";
		BdryFormat=0;				// type of boundary condition we are applying
									// 0 = constant value of A
									// 1 = Small skin depth eddy current BC
									// 2 = Mixed BC

		A0=0.; A1=0.;
		A2=0.; phi=0.;			// set value of A for BdryFormat=0;
	
		Mu=0.; Sig=0.;			// material properties necessary to apply
								// eddy current BC
		
		c0=0.; c1=0.;			// coefficients for mixed BC

}

CPointProp::CPointProp()
{
		PointName="New Point Property";
		Jr=0.; Ji=0.;					// applied point current, A 
		Ar=0.; Ai=0.;					// prescribed nodal value;
}

CCircuit::CCircuit()
{
		CircName="New Circuit";
		dVolts=0.;
		Amps=0.;
		CircType=0;
		J=0.;
		Case=0;
}